
/*
   Virtual Machine Ropeland
   
   @TITLE Rope Land
   @AUTHOR R.Corda
   @VERSION 0.4.21.11052000

   Copyright (c) 2000, Roberto Corda (roberto.corda@inferentia.it)
   All rights reserved.

   This file is subject to the terms of the GNU General Public License as
   published by the Free Software Foundation.  A copy of this license is
   included with this software distribution in the file COPYING.  If you
   do not have a copy, you may obtain a copy by writing to the Free
   Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

//fix: separare la parte di input dalla parte di pura simulazione
//fix: questi aggiunti per keypressed() e kbhit()
#include <allegro.h>
#include <pc.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "rope.h"
#include "ropeasm.h"

#define GROUP "ACME inc 2000"
#define TITLE "Rope Land"
#define AUTHOR "Roberto Corda"
#define VERSION "v0.4.21.15052000"

#define OK 0

// errori esecuzione comandi (OK == 0)
// errori per dup
enum cmderrs { DEAD=1,NOPLACE,NOBODY,OUTOFBOUND,NOERG,
               CONTACT,PUSH,BOUNCE,DAMAGE };

//livelli di errore
#define ERROR 1
#define WARN 2
#define SILENT 3

//quantita' di dati satmpati insieme all'errore
#define NODATA 0
#define DATA 1
#define NOPRODATA 2

// tipi di segnale
#define NOSIGNAL 0
// ogni quanto il segnale viene controllato e cancellato se vecchio
#define DELSPKSTEP 10
// energia presa con un *morso* L == TREE, V == LAMB, M = LION
// fix: metterle in variabili per poterle cambiare a runtime?
#define BYTERGL 5
#define BYTERGV 40
#define BYTERGM 150

//tipo di scontro
#define MATCHTY 0
#define DARWINTY 1
#define DEBUGTY 2
#define DEBUGMATY 3
#define DEBUGDATY 4
#define MATCHNOGTY 5

#define _decerg(o) o->erg--
#define adderg(o,v) o->erg += v

#define getpc(o) (o->proc->status.pc)
#define getid(o) (o->id)
#define gettype(o) (o->type)
#define _geterg(o) (o->erg)
#define getweight(o) (o->weight)
#define getfather(o) (o->father_id)
#define getdir(o) (o->dir)
#define getbirth(o) (o->birthday)
#define getx(o) (o->x)
#define gety(o) (o->y)
#define getcmd(o,pc) (o->proc->cmd[pc])
#define getnumcmd(o) (o->proc->numcmd)
#define getrlm(o,pc) (o->proc->cmd[pc].rlm)
#define getp1(o,pc) (o->proc->cmd[pc].p[0])
#define getp2(o,pc) (o->proc->cmd[pc].p[1])
#define getp3(o,pc) (o->proc->cmd[pc].p[2])
#define _getvalue(o,i) (o->proc->status.vars[i])
#define getsignal(x,y) (ropeland[x][y].signal)

#define getrlmname(o,pc) (rlmname[o->proc->cmd[pc].rlm])

#define putid(o,idf) o->id = idf
#define putvalue(o,i,v) o->proc->status.vars[i] = v
#define _putpc(o,v) o->proc->status.pc = v
#define putx(o,v) o->x = v
#define puty(o,v) o->y = v
#define _putsignal(x,y,w) ropeland[x][y].signal = w
#define puttime(x,y) ropeland[x][y].time

#define isalive(o) (o->alive)
#define isa(o,t)(o->type == t)
#define isindexp1(o,pc) (o->proc->cmd[pc].ctrl[0])
#define isindexp2(o,pc) (o->proc->cmd[pc].ctrl[1])
#define isindexp3(o,pc) (o->proc->cmd[pc].ctrl[2])

char *typesname[6] = {"rock","tree","lamb","lion","sign","bkgr"};
int deltax[8] = {0,1,1,1,0,-1,-1,-1};
int deltay[8] = {-1,-1,0,1,1,1,0,-1};
char *ctrlname[2] = {"#"," "};
char *rlmname[POOLCMD] =
     {"notacmd","load", "if", "halt",
     "and","or","not","lt","eq",
     "move", "speak", "ear", "see",
     "dup","eatv", "eatm", "eatl", "feed",
     "add","sub","inc","dec",
     "radar", "erg","father","type","id","birth","tic","rand"};
int  npar[POOLCMD] =
     {0,2,2,0,
      3,3,1,3,3,
      3,2,1,2,
      3,2,2,1,2,
      3,3,1,1,
      2,1,1,1,1,1,1,2} ;
//sub dir
int sdir[8][3] = {{NW,NORTH,NE},
                  {NORTH,NE,EAST},
                  {NE,EAST,SE},
                  {EAST,SE,SOUTH},
                  {SE,SOUTH,SW},
                  {SOUTH,SW,WEST},
                  {SW,WEST,NW},
                  {WEST,NW,NORTH}
                 };
//direzione dopo uno scontro (rimbalzo)
//inverte, scontro contro parete verticale, orizzontale
int ibounce[8] = {SOUTH,SW,WEST,NW,NORTH,NE,EAST,SE};
int vbounce[8] = {NORTH,NW,WEST,SW,SOUTH,SW,EAST,NE};
int obounce[8] = {SOUTH,SE,EAST,NE,NORTH,NW,WEST,SW};
//uin funzione della direzione in cui procede direzioni
//in cui e' possibile ci siano le entita' contro cui puo' rimbalzare
int hitdir[8][2] = {{SOUTH,SOUTH},
                    {EAST,NORTH},
                    {SOUTH,SOUTH},
                    {EAST,SOUTH},
                    {SOUTH,SOUTH},
                    {WEST,SOUTH},
                    {SOUTH,SOUTH},
                    {WEST,NORTH}};

int start(tobj *obj);
tobj* initobj (int type,int erg,int weight);
int putcmd(tobj *obj, int rlm, char c1,int p1,
                char c2,int p2,char c3,int p3);
int dummyprog(tobj *obj);
int genesis();
int newid();
int start(tobj *obj);
int kill(tobj *obj);
int steprun(int schindex);
int error(tobj *obj,int data,int level,int cond,char *comment);
int addobj(int father_id,int type,int erg,int weight,
           int x,int y,tcmd *prog,int numcmd);

void inerziamove(void);
int suberg(tobj *obj, int delta);
// builtins
int speak(tobj *obj, int word, int erg);
//void load(tobj *obj,int ivar, int value);
void putallgui();
void  _calcmove(int x,int y,int *px,int *py,int dir);

//variabili esterne da ropegui
extern int nograph;

//variabili globali
tland ropeland[LATO+2][LATO+2];  // terreno di gioco
tobj *ropes[MAXROPES];           // ropes viventi sassi e cadaveri
int id;                          // ultimo id assegnato
int active;                      // numero oggetti schedulati
int alive;                       // numero di oggetti vivi
long tic;                        // tempo passato
int type;                        // tipo di scontro MATCH o DARWIN
int debugval;                    // se vero serve per il debug

tcmd t[MAXCMD];                  // comandi caricati se tipo debug
int n;                           // numero comandi



//cancella i segnali scaduti
void delsignal(void) {
 int i,j;

 for (i=0; i<LATO + 2; i++)
   for (j=0; j<LATO + 2; j++) {
   if (ropeland[i][j].signal != NOSIGNAL)
     if (tic - ropeland[i][j].time >= DELSPKSTEP) {
       ropeland[i][j].signal = NOSIGNAL;
       if ((ropeland[i][j].obj == NULL) && (!debug()))
         putgui(i,j, BKG);
     }
   }
}

// sistema operativo VM
int god() {
  int a;

  for (a=0; a<active; a++){
    steprun(a);
    // ogni DELSPKSTEP controlla se ci sono segnali da cancellare
    if (tic%DELSPKSTEP == 0)
      delsignal();
  }
  //move oggetti per la loro inerzia (energia.cinetica)
  inerziamove();

  return active && alive;  // deve essercene almeno uno
                           // attivo che deve essere vivo
}

// calcola la direzione di rimbalzo
int bounce(int x, int y, int dir) {
 int xo,yo,xv,yv,exista,existb;

 dir = dir%8;

 //direzione opposta
 if ((dir == NORTH) || (dir ==EAST) || (dir == SOUTH) || (dir == WEST))
   return ibounce[dir];

 //rimbalzo con angolo
 _calcmove(x,y, &xo, &yo, hitdir[dir][0]);
 _calcmove(x,y, &xv, &yv, hitdir[dir][1]);

 //esistenza di elementi vicino al punto di scontro
 //aka parete verticale o orizzontale
 //fix: se il nuovo elemento non e' amovibile dovrebbeforse
 //fix: muoversi lui
 exista = (ropeland[xo][yo].obj != NULL);
 existb = (ropeland[xv][yv].obj != NULL);

 //contro una parete verticale
 if (exista && !existb)
   return vbounce[dir];
 //contro una parete orizzontale
 if (!exista && existb)
   return obounce[dir];
 //torna indietro
 return ibounce[dir];
}

int main(int argc,char *args[]){
  int pos;
  int nf;

  // seme per rand
  srand(time(0));

  //di default non e' in stato di debug
  debugval = FALSE;

  //di default il tipo di scontro e' MATCH
  type = MATCHTY;

  genesis();

  if ((args[1] == NULL)) {
    printf("usage: ropeland [-match | -darwin | -debug | -nograph] <file.inc>\n\r");
    printf("       ropeland [-single] <file.asm>\n\r");
    printf("       ropeland [-ver]\n\r");

    return 0;
  }

  if (!strcmp(args[1],"-ver")) {
    printf("   Copyright (c) 2000, %s\n\r",AUTHOR);
    printf("         All rights reserved.\n\r");
    printf("     GNU General Public License.\n\r\n\r");
    printf("   %s version %s\n\r",TITLE,VERSION);
    //printf("task group %s",GROUP);
    return FALSE;
  } else if (!strcmp(args[1],"-match") || !strcmp(args[1],"-nograph")) {
    if (!strcmp(args[1],"-nograph")) {
      nograph = TRUE;
      printf ("graph mode OFF\n\r");
    } else   {
      nograph = FALSE;
      printf ("graph mode ON\n\r");
    }
    type = MATCHTY;
    if (args[2]==NULL) {
      printf("usage: ropeland -match <file.inc>\n\r");
      return 0;
    }
  } else if (!strcmp(args[1],"-darwin")) {
    type = DARWINTY;
    printf("Sorry: -darwin mode not yet implemented\n\r");
    if (args[2]==NULL) {
      printf("usage: ropeland -darwin <file.inc>\n\r");
      return 0;
    }
    exit(0);
  } else if (!strcmp(args[1],"-single")){
     printf("Acme_inc DEBUG MODE\r\n");
     debugval = TRUE;
     type = DEBUGTY;
     if (args[2]==NULL) {
      printf("usage: ropeland -single <file.asm>\n\r");
      return 0;
     }
  } else if (!strcmp(args[1],"-debug")) {
     printf("Acme_inc DEBUG MATCH\r\n");
     debugval = TRUE;
     type = MATCHTY;
     if (args[2]==NULL) {
      printf("usage: ropeland -debug <file.inc>\n\r");
      return 0;
     }
  }

  if (type == DEBUGTY) {
    n = loadsource(args[2],t);
    if (!n)
      exit(0);
    printf("DEBUG MODE [fixit]: %s is a lamb\n\r",args[2]);
    addobj(GOD,LAMB,1170,42,10,3,t,n);
  } else {
    if (!loadinc(args[2]))
      exit(0);
  }

  /*
   * Inizio simulazione
  */
  if (!debug()) {
   initgui();
   putallgui();
  }

  while (nograph ? (!kbhit()) : (!keypressed()))
      if (!god()) break;

  // fine simulazione
  if (!debug())
    closegui();

  printf("bye!\n\r");

  return 0;
}

void putallgui() {
 int i;
 for (i=0; i<active; i++){
   putgui(getx(ropes[i]),gety(ropes[i]),gettype(ropes[i]));
 }
}

// mette nello schedulatore il nuovo oggetto
// restituisce posizione in ropes[]
// father_id == GOD significa immessi dall'esterno
int addobj(int father_id,int type,int erg,int weight,
           int x,int y,tcmd *prog,int numcmd){
  tobj *obj;
  int i;
  obj = initobj(type,erg,weight);
  error(obj,NODATA,ERROR,active >= MAXROPES,"addobj, too many objs");
  ropes[active] = obj;
  if (!isa(obj,ROCK))  // le pietre non sono vive
    alive++;
  place(obj, x,y);      // mette il rope sul terreno di gioco
  start(obj);           // inizializza il processo

  // aggiunge programma aka insieme programmi
  for (i=0;i<numcmd;i++){
    obj->proc->cmd[i] = prog[i];
  }
  obj->proc->numcmd = numcmd;

  // all'inizio e' fermo
  obj->dir = 0;
  obj->inerzia = 0;

  return(active++);     // numero posizione in ropes
}

// carica il valore value nella posizione in memoria ivar
void load(tobj *obj,int ivar, int value) {
  error(obj,DATA,ERROR,(ivar<0) || (ivar >= MAXVARS),
    "load, out of memory bounds");
  putvalue(obj, ivar, value);
}

// toglie oggetto da x,y in ropeland
int unplace(int x,int y) {
  ropeland[x][y].obj = NULL;
  return TRUE;
}

// rimette un oggetto in x,y di ropeland
int replace(tobj *obj,int x,int y) {

  ropeland[x][y].obj = obj;
  return TRUE;
}

void getpos(tobj *obj,int *px,int *py){
  *px = getx(obj);
  *py = gety(obj);
}

//calcola la nuova posizione procedendo di una casella nella direzione dir
void  _calcmove(int x,int y,int *px,int *py,int dir) {
  //fix dir ciclica (oppure restituire un errore)
  dir = dir%8;
  *px = x + deltax[dir];
  *py = y + deltay[dir];
}

//come _calcmove, mette risultato nell'oggetto
void calcmove(tobj *obj,int *px,int *py,int dir) {
  _calcmove(getx(obj),gety(obj),px,py,dir);
}

//muove effettivamente gli oggetti
void inerziamove(void) {
 int a;
 int x1,x2,y1,y2;

 for (a=0; a<active; a++){
  if (ropes[a]->inerzia) {
    if (!move(ropes[a],ropes[a]->dir,0)) {
      getpos(ropes[a],&x1,&y1);
      calcmove(ropes[a], &x2, &y2, ropes[a]->dir);
      _move(ropes[a],x1,y1,x2,y2);
      movegui(x1,y1,x2,y2,ropes[a]->type);
      // l'energia persa per attrito fissa == 10, variabile?
      if ((ropes[a]->inerzia -= 10) < 0)
        ropes[a]->inerzia = 0;
    }
  }
 }
}

//move oggetto senza controlli nella struttura dati
int _move(tobj *obj, int x1, int y1, int x2, int y2) {
   unplace(x1,y1);
   replace(obj,x2,y2);
   putx(obj,x2);
   puty(obj,y2);
   return TRUE;
}
// toglie unquanto di energia all'oggetto e se non e' morto
// restituisce il valore in out dato
int outsuberg(tobj *obj,int erg, int out) {
  if (suberg(obj,erg))
     return DEAD; // e' morto dallo sforzo
  else
     return out;
}

// muove un oggetto in una certa direzione
// restituisce != 0 se non si muove per qualche ragione
// DEAD se muore
// CONTACT se ha toccato qualcosa
// NOERG se non abbastanza energia
// PUSH se spinge un altro elemento
// DAMAGE se danneggia un altro elemento
// BOUNCE se rimbalza via
int move(tobj *obj, int dir, int erg) {
  int x1,x2,y1,y2;
  int out; // valore restituito
  int p,toterg; // peso, inerzia piu' energia aggiunta

  //fix: diminuzione dell'energia fornita all'inerzia
  //fix: caso DEAD
  //aggiunge energia all'inerzia
  obj->inerzia += erg;
  obj->dir = dir;
  toterg= obj->inerzia;

  if (error(obj,DATA,WARN,dir >= 8,"move, dir dosn't exist"))
    dir = dir%8;
  //calcola posizione presente e successiva (x1,y1) -> (x2,y2)
  getpos(obj,&x1,&y1);
  calcmove(obj, &x2, &y2, dir);

  if (error(obj,DATA,WARN,(x2<0) || (y2<0) || (x2>LATO+1) || (y2>LATO+1),
       "move, out of bounds"))
     return outsuberg(obj,erg,OUTOFBOUND);

   p = 10; //fix: peso di questo oggetto temporaneamente costante

   // ropeland[x2][y2].obj.erg / 10;
   // fix: funzione (energia) --> peso
  
   // se ha un oggetto davanti
   if (error(obj,DATA,WARN,ropeland[x2][y2].obj != NULL,"move, contact")){

    // se l'oggetto e' inamovibile rimbalza
    if (ropeland[x2][y2].obj->erg == 0)  {
       obj->dir =  bounce(getx(obj), gety(obj), getdir(obj));
       return outsuberg(obj,erg,BOUNCE);
    }

    //spostamenti di altri oggetti
    p = 10; // peso dell'altro oggetto fix: temporaneamente costante

    // se l'energia non e' sufficiente a spostarlo
    if (toterg < p)  {
      //se si scontra contro una roccia questo rimbalza
      if (ropeland[x2][y2].obj->type == ROCK) {
         obj->dir = bounce(x1,y1, dir);
         return outsuberg(obj,erg,BOUNCE);
      } else {
          //fix: qui mancano i danni all'altro oggetto se biologico;
          obj->inerzia = 0;
          return outsuberg(obj,erg,DAMAGE);
      }
    } else {
      //energia sufficiente a spostare l'altro
      ropeland[x2][y2].obj->inerzia += toterg;
      obj->inerzia = 0;
      ropeland[x2][y2].obj->dir = dir;
      return outsuberg(obj,erg,PUSH);
    }
  }

   //non c'e' stato contatto
   if (toterg >= p)
     return outsuberg(obj,erg,OK);
   else
     return outsuberg(obj,erg,NOERG);
}

// duplica il rope nella direzione indicata
// concedendogli l'energia indicata
// fix: stabilire quanta energia e' necessaria
// fix: alla duplicazione in se
// restituisce OK == 0 se andato a buon fine
// DEAD se morto
int dup(tobj *obj,int dir, int erg) {
  int x,y,i;
  int pos;

  // in ogni caso un tot di energia viene usata fix: quanta?
  if (suberg(obj,20))
     return DEAD; // e' morto di parto
  //fix: la direzione potrebbe essere sbagliata
  //fix: error(dir<0 || dir>=8?)
  //fix messa ciclica 8 numero di direzioni dentro calcmove
  calcmove(obj,&x,&y,dir);

  if (error(obj,DATA,WARN,ropeland[x][y].obj != NULL,
    "dup, no room for a baby")){
    return NOPLACE;
  }

  error(obj,DATA,ERROR,(x<0) || (y<0) || (x>LATO+1) || (y>LATO+1),
       "dup, out of bounds");
  // fix: non e' detto che il padre abbia la necessaria energia da
  // fix: passare al figlio
  // fix: peso messo a caso: 42
  // il padre cede parte della sua energia al figlio
  pos = addobj(getid(obj),gettype(obj),erg,42,
               x,y,obj->proc->cmd,getnumcmd(obj));
  //disegna nuovo oggetto
  putgui(x,y,gettype(obj));
  return (suberg(obj,erg));
}

// mangia un TREE togliendogli BYTERGV energia
// restituisce valore > 0 se ha avuto successo
int eatv(tobj *obj, int dir) {
  int x,y;
  tobj *target;

  calcmove(obj,&x,&y,dir);
  target = ropeland[x][y].obj;
  if (error(obj,DATA,WARN,target == NULL,
     "eatv, nothing there"))
     return 0;
  if (error(obj,DATA,WARN,!isa(target,TREE),
     "eatv, lamb eat tree"))
     return 0;
  // fix: l'energia richiesta puo' anche non essere presente
  // fix: quanto e' energetica l'erba?
  suberg(target,BYTERGV);
  adderg(obj,BYTERGV);
  return BYTERGV;
}

// mangia un LAMB o un LION togliendogli BYTERGM energia
// restituisce valore > 0 se ha avuto successo
// fix: da unificare con eatv (quasi uguali)
int eatm(tobj *obj, int dir) {
  int x,y;
  tobj *target;

  calcmove(obj,&x,&y,dir);
  target = ropeland[x][y].obj;
  if (error(obj,DATA,WARN,target == NULL,
     "eatm, nothing there"))
     return 0;
  if (error(obj,DATA,WARN,isa(target,TREE),
     "eatm, lion eat lamb or lion"))
     return 0;
  // fix: l'energia richiesta puo' anche non essere presente
  suberg(target,BYTERGM);
  adderg(obj,BYTERGM);
  return BYTERGM; // restituisce valore > 0 se ha avuto successo
}

//come eatv e eatm, mangia luce (prende energia dall' *esterno*)
int eatl(tobj *obj) {
  if (error(obj,DATA,WARN,!isa(obj,TREE),
     "eatl, only trees eat light"))
     return 0;
  adderg(obj,BYTERGL);
  return BYTERGL;
}

//fix: non testato
//passa dell'energia ad un entita' contigua
int feed(tobj *obj, int dir, int erg) {
  int x,y;
  tobj *target;

  target = ropeland[x][y].obj;
  calcmove(obj,&x,&y,dir);

  if (error(obj,DATA,WARN,target == NULL,
    "feed, nobody there"))
    return NOBODY;
  if (error(obj,DATA,WARN,isa(target,ROCK),
    "feed, cannot feed a stone"))
    return NOBODY;

  adderg(target,erg);
  return suberg(obj,erg);
}

//aggiunge un comando al programma di una entita'
int putcmd(tobj *obj, int rlm, char c1,int p1, char c2,int p2,char c3,int p3) {
  int nc = getnumcmd(obj);
  error (obj,NODATA,ERROR,
    obj->proc == NULL,"putcmd, process is NULL");
  error (obj,DATA, ERROR,
    nc >= MAXCMD,"putcmd, too many commands");

  //fix: usare le #define adatte
  obj->proc->cmd[nc].rlm = rlm;

  obj->proc->cmd[nc].p[0] = p1;
  obj->proc->cmd[nc].ctrl[0] = c1;
  obj->proc->cmd[nc].p[1] = p2;
  obj->proc->cmd[nc].ctrl[1] = c2;
  obj->proc->cmd[nc].p[2] = p3;
  obj->proc->cmd[nc].ctrl[2] = c3;

  return obj->proc->numcmd++;
}

// inizializza la struttura dati ropeland[][]
int genesis() {
  int i,j;

  //init ropeland
  // crea il silenzio
  for(i=0;i<LATO+2;i++)
  for(j=0;j<LATO+2;j++)
  {
   ropeland[i][j].obj = NULL;
   ropeland[i][j].signal = NOSIGNAL;
  }

  //init schedulatore
  for(i=0;i<MAXROPES;i++)
   ropes[i] = NULL;

  id = 0;      // max id
  tic = 0;     // tempo passato
  active = 0;  // numero di oggetti attivi
  alive = 0;   // numero di oggetti vivi

  //aggiunge massi sui lati del terreno di gioco
  for(i=0;i<LATO+2;i++) {
    addobj(GOD,ROCK,0,1000,i,0,NULL,0);
    addobj(GOD,ROCK,0,1000,i,LATO+1,NULL,0);
  }
  for(i=1;i<=LATO;i++) {
    addobj(GOD,ROCK,0,1000,0,i,NULL,0);
    addobj(GOD,ROCK,0,1000,LATO+1,i,NULL,0);
    // fix: energia zero significa inamovibile per le ROCK
  }
}

//crea un nuovo identificatore di ropes
int newid() {
  id++;
  return id;
}

//posiziona nuovo oggetto sul terreno di gioco
int place(tobj *obj, int x, int y) {
  putx(obj,x);
  puty(obj,y);
  error(obj,DATA,ERROR,(x<0) || (y<0) || (x>LATO+1) || (y>LATO+1),
       "place, out of bounds");
  ropeland[x][y].obj = obj;
}

//inizializza i valori del nuovo oggetto
tobj* initobj (int type,int erg,int weight) {
  tobj *obj;

  obj = (tobj *)malloc(sizeof(tobj));
  putid(obj, newid());
  obj->type = type;
  if (type==ROCK)
    obj->alive = FALSE;
  else
    obj->alive = TRUE;
  obj->proc = NULL;
  obj->erg = erg;
  obj->weight = weight;

  return obj;
}

//inizializza oggetto per essere eseguito
int start(tobj *obj){
 obj->birthday = tic;
 if(error(obj, NODATA, SILENT,isa(obj,ROCK),
   "start, no processes for stones")) {
   obj->alive = FALSE;
   obj->proc = NULL;
   return;
 }
 obj->alive = TRUE;
 obj->proc = malloc(sizeof(tprocess));
 obj->proc->status.pc = 0;
 obj->proc->numcmd = 0;
}

// uccide oggetto senza toglierlo dallo schedulatore
// e dal terreno di gioco (cadavere)
int kill(tobj *obj){
 obj->alive = FALSE;
 free(obj->proc);
 obj->proc = NULL;
 alive--;
}

//toglie l'oggetto dallo schedulatore e dal terreno di
//gioco, ritorna numero di oggetti attivi
int destroy(const int schindex) {
  tobj *obj = ropes[schindex];
  unplace(getx(obj),gety(obj)); // toglie da schermo
  free(obj);
  ropes[schindex] = NULL;       // toglie da schedulatore
  if (active>1){
    ropes[schindex] = ropes[active-1];
    ropes[active-1] = NULL;    // sposta l'ultimo nella posizione vuota
  }
  active--;
  return active;
}

//fix: da mettere in un file common.c con tutte le funzioni
//fix: di utilita' generale che non dipendono dalla struttura dati
//determina il livello di debug del sistema
int debug() {
 //fix: potrebbero esserci vari livelli di debug dove
 //fix: 0 == NODEBUG
 return debugval;
}

//stampa un comando su stdout
void printcmd(tobj *obj, int pc) {
  int rlm;
  if (error(obj,NODATA,ERROR, obj->proc == NULL,
    "printcmd, cannot print NULL process"))
    return;

  rlm = getrlm(obj,pc);
  printf("%-5d:(%2d:%s)(x:%2d y:%2d) %2d: ", tic, getid(obj),
    typesname[gettype(obj)],
    getx(obj),gety(obj),pc, getrlmname(obj,pc));

  //_printcmd
  // fix: cambiare getp#() in modo passargli l'indice
  // fix: in modo da abbreviare questo
  printf("%s ",getrlmname(obj,pc));
  if (npar[rlm] >= 1) {
    if (isindexp1(obj,pc))
      printf("(%d)*",getvalue(obj,getp1(obj,pc)));
    printf("%d ",getp1(obj,pc)) ;
  if (npar[rlm] >= 2) {
    if (isindexp2(obj,pc))
      printf("(%d)*",getvalue(obj,getp2(obj,pc)));
    printf("%d ",getp2(obj,pc)) ;
  if (npar[rlm] >= 3) {
    if (isindexp3(obj,pc))
      printf("(%d)*",getvalue(obj,getp3(obj,pc)));
    printf("%d ",getp3(obj,pc)) ;
    }
   }
  }

  printf("\n\r");
}

//restituisce TRUE se e' avvenuto l'errore ovvero se cond == 1
//stampa stato del sistema a seconda del valore di level e data
int error(tobj *obj,int data,int level,int cond,char *comment){

  if (cond) {

   if (level != SILENT) {
   if (debug() || (level == ERROR)) {
    switch (level) {
      case ERROR:
        printf("FATAL:");
        break;
      case WARN:
        printf("WARNING:");
        break ;
      default:
        error(obj,NODATA,ERROR,1,"error, unknow level");
     }

    switch (data) {
      case NODATA:
        printf("%s (NODATA)\n\r",comment);
        break;
      case DATA:
        printf("%s (id:%d pc:%d ty:%s er:%d we:%d x:%d y:%d)\n\r",
        comment,getid(obj),getpc(obj),typesname[gettype(obj)],
          _geterg(obj),getweight(obj),getx(obj),gety(obj)) ;

        break;
      case NOPRODATA:
        printf("%s (id:%d ty:%s er:%d we:%d x:%d y:%d)\n\r",
          comment,getid(obj),typesname[gettype(obj)],
          _geterg(obj),getweight(obj),getx(obj),gety(obj));
        break;
     }
    }
   }
   if (level == ERROR)
     exit(0);
 }
 return cond;
}

//cambia il valore del program counter (pc) aka fa un goto
void putpc(tobj *obj, int pos){
  error(obj,DATA,ERROR,pos >= getnumcmd(obj),"putpc, jump out of bound");
  _putpc(obj,pos);
}

//incrementa il program counter
void incpc(tobj *obj) {
  putpc(obj, getpc(obj)+1);
}

// decrementa l'energia dell'oggetto
// se muore restituisce DEAD
int decerg(tobj *obj){
  if (error(obj,NOPRODATA,WARN, _geterg(obj) <= 0,"decerg, no energy left")){
    kill(obj);
    return DEAD;
  } else
    _decerg(obj);
  return OK;
}

// decrementa l'energia dell'oggetto
// se muore restituisce DEAD
int suberg(tobj *obj, int delta){
  if (error(obj,DATA,WARN, (_geterg(obj) <= 0) || (obj->erg <= delta),
    "suberg, no energy left")) {
    obj->erg = 0;
    kill(obj);
    return DEAD;
  }
  obj->erg -= delta;
  return OK;
}

//restituisce il valore contenuto nella posizione index
int getvalue(tobj *obj, int index) {
  error(obj,DATA,ERROR, (index<0) || (index >= MAXVARS),
    "getvalue, index out of bound");
  return _getvalue(obj,index);
}

// guarda nella direzione (dir,subdir) restituisce il tipo della
// prima entita' incontrata
int see(tobj *obj, int dir, int subdir){
  int x,y;

  if (error(obj,DATA,WARN,dir >= 8,"see, dir dosn't exist"))
    dir = dir%8; //fix: dir ciclica se non esistente
                 //fix: oppure return? di cosa?

  //calcola le posizioni iniziali delle linee visuali
  calcmove(obj,&x,&y,sdir[dir][subdir]);
  //fix: presume che incontri sempre qualcosa prima di uscire dai bordi
  do {
   //se trova qualcosa esce
   if (ropeland[x][y].obj != NULL)
     return gettype(ropeland[x][y].obj);

   // altrimenti avanza
   _calcmove(x,y,&x,&y,dir);
  } while(1);
}


// restituisce la distanza del primo oggetto incontrato lungo la direzione
// e sottodirezione
int radar(tobj *obj, int dir,int subdir){
  int x,y;
  int out=0;

  if (error(obj,DATA,WARN,dir >= 8,"radar, dir dosn't exist"))
    dir = dir%8; //fix: dir ciclica se non esistente
                 //fix: oppure return? di cosa?
  //calcola le posizioni iniziali delle linee visuali
  calcmove(obj,&x,&y,sdir[dir][subdir]);
  //fix: presume che incontri sempre qualcosa prima di uscire dai bordi
  do {
   //se trova qualcosa esce
   if (ropeland[x][y].obj != NULL) return out;

   // altrimenti avanza
   calcmove(obj,&x,&y,dir);
   out++;
  } while(1);
}

// mette il segnale word in posizione x,y in ropeland
// associandoci il tempo attuale (i segnali decadono dopo un certo tempo)
// restituisce TRUE se fuori dal campo di gioco
int putsignal(int x,int y,int word) {
  // fix: mettere un error(..) in modo SILENT
  if ((x<0) || (y<0) || (x>=LATO+2) || (y>=LATO+2))
    return TRUE;
  _putsignal(x,y,word);
  puttime(x,y);

  if ((ropeland[x][y].obj == NULL) && // non copre oggetti presenti
       (!debug()))
    putgui(x,y,SIGNAL);
  return FALSE;
}

// aggiunge a spirale attorno all'oggetto il segnale word
// si ferma quando finisce l'energia (1 erg per ogni casella con segnale)
// ritorna OK se tutto a posto DEAD se morto per l'urlo
// fix: appena esce dal campo di gioco smette di propagare il suono
// fix: il suono vero ha anche una direzione
int speak(tobj *obj, int word, int erg) {
 int x,y,x1,y1;
 int dx,dy,l;

 x = getx(obj);
 y = gety(obj);
 for(l=3;1;l+=2){
   _calcmove(x,y,&x,&y,NW);
   for (dx = 0;dx<l;dx++) {
     if (putsignal(x+dx,y,word))
       return OK;
     if (decerg(obj))
       return DEAD; // e' morto per l'urlo
     if (--erg == 0)
       return OK;
   }
   for (dy = 1;dy<l;dy++) {
     if (putsignal(x+l-1,y+dy,word))
       return OK;
     if (decerg(obj))
       return DEAD; // e' morto per l'urlo
     if (--erg == 0)
       return OK;
   }
   for (dx = l-2;dx>=0;dx--) {
     if (putsignal(x+dx,y+l-1,word))
       return OK;
     if (decerg(obj))
       return DEAD; // e' morto per l'urlo
     if (--erg == 0)
       return OK;
   }
   for (dy = l-2;dy>0;dy--) {
     if (putsignal(x,y+dy,word))
       return OK;
     if (decerg(obj))
       return DEAD; // e' morto per l'urlo
     if (--erg == 0)
       return OK;
   }
 }

}
// processa una riga di comando dell'oggetto di indice
// schindex in rope
// ritorna DEAD se morto OK se tutto e' andato bene
// OK == 0
int steprun(int schindex) {
 int pc,rlm,p1,p2,p3; //parametri
 int buff;

 tobj *obj = ropes[schindex];

 if (error(obj, NOPRODATA, SILENT, !isalive(obj),
   "steprun, obj is dead")) {
   if (!isa(obj,ROCK)) {
     if (_geterg(obj) <= 0) {
       delgui(getx(obj),gety(obj));  // lo toglie dallo schermo
       destroy(schindex);    // se e' morto con energia 0 lo distrugge
     }
     else
       decerg(obj);          // altrimenti lo fa deperire
   }
   return DEAD;
 }
 error(obj, NODATA, ERROR, obj->proc == NULL,
   "steprun, process is null");

 pc = getpc(obj);
 rlm = getrlm(obj,pc);
 // fix: dovrebbe caricare i parametri in funzione
 // fix: del numero di parametri del comando

 error(obj,DATA,ERROR, (rlm<0) || (rlm>=POOLCMD),
    "steprun, not a command");

 if (npar[rlm] >= 1) {
 p1 = getp1(obj,pc);
 if (isindexp1(obj,pc))
   p1 = getvalue(obj,p1);
 }

 if (npar[rlm] >= 2) {
 p2 = getp2(obj,pc);
 if (isindexp2(obj,pc))
   p2 = getvalue(obj,p2);
 }

 if (npar[rlm] >= 3) {
 p3 = getp3(obj,pc);
 if (isindexp3(obj,pc))
   p3 = getvalue(obj,p3);
 }

 error(obj, DATA, ERROR, pc >= getnumcmd(obj),
   "steprun, pc out of bound");
 if (debug())
   printcmd(obj,pc);
 switch (rlm) {

   // comandi di controllo processo
   case LOAD:
     load(obj,p1,p2);
     incpc(obj);
     break;   
   case IF:
     if (p1)
       putpc(obj,p2); //p2 viene considerato un value non un ivar
     else
       incpc(obj);
     break;
   case HALT:
     kill(obj);
     return DEAD;
     break;

   //comandi di confronto
   case AND:
     load(obj,p1,p2&&p3);
     incpc(obj);
     break;
   case OR:
     load(obj,p1,p2||p3);
     incpc(obj);
     break;
   case NOT:
     load(obj,p1,!getvalue(obj,p1));
     incpc(obj);
     break;
   case LT:
     load(obj,p1,p2<p3);
     incpc(obj);
     break;
   case EQ:
     load(obj,p1,p2==p3);
     incpc(obj);
     break ;
   
   //comandi di I/O
   case MOVE:
     if (error(obj,DATA,ERROR,isa(obj,TREE),
       "steprun, tree doesn't move"))
       return FALSE;
     buff = move(obj,p2,p3);
     if (buff == DEAD)
       return DEAD;
     load(obj,p1,buff);
     incpc(obj);
     break;
   case SPEAK:
     if (error(obj,DATA,ERROR,isa(obj,TREE),
       "steprun, tree doesn't speak"))
       return 0;
     //fix: a che serve buff? ricontrollare
     buff = speak(obj,p1,p2);

     // se muore per l'urlo non continua l'operazione
     if(buff == DEAD)
       return DEAD;

     incpc(obj);
     break;
   case EAR:
     if (error(obj,DATA,ERROR,isa(obj,TREE),
       "steprun, tree doesn't ear"))
       return 0;
     load(obj,p1,getsignal(getx(obj),gety(obj)));
     incpc(obj);
     break;
   case SEE:
     if (error(obj,DATA,ERROR,isa(obj,TREE),
       "steprun, tree doesn't see"))
       return 0;
     load(obj,p1,see(obj,p2,0));
     load(obj,p1+1,see(obj,p2,1));
     load(obj,p1+2,see(obj,p2,2));
     incpc(obj);
     break;
   case RADAR:
     if (error(obj,DATA,ERROR,isa(obj,TREE),
       "steprun, tree doesn't see"))
       return 0;
     load(obj,p1,radar(obj,p2,0));
     load(obj,p1+1,radar(obj,p2,1));
     load(obj,p1+2,radar(obj,p2,2));
     incpc(obj);
     break;
   
   // comandi biologici
   // NB (in corewars il comando dup non esiste)
   case DUP:
     buff = dup(obj,p2,p3);

     // se muore di parto non continua l'operazione
     if(buff == DEAD)
       return DEAD;

     load(obj,p1,buff);
     incpc(obj);
     break;
   case EATV:
     load(obj,p1,eatv(obj,p2));
     incpc(obj);
     break;
   case EATM:
     load(obj,p1,eatm(obj,p2));
     incpc(obj);
     break;
   case EATL:
     load(obj,p1,eatl(obj));
     incpc(obj);
     break;
   case FEED:
     buff = feed(obj,p1,p2);
     if (buff == DEAD)
        return DEAD;
     incpc(obj);
     break;
   case ADD:
     load(obj,p1,p2+p3);
     incpc(obj);
     break;
   case SUB:
     load(obj,p1,p2-p3);
     incpc(obj);
     break;
   case INC:
     load(obj,p1,getvalue(obj,p1)+1);
     incpc(obj);
     break;
   case DEC:
     load(obj,p1,getvalue(obj,p1)-1);
     incpc(obj);
     break;
     //comandi di sistema
   case ERG:
     load(obj,p1,_geterg(obj));
     incpc(obj);
     break;
   case FATHER:
     load(obj,p1,getfather(obj));
     incpc(obj);
     break;
   case TYPE:
     load(obj,p1,gettype(obj));
     incpc(obj);
     break;
   case ID:
     load(obj,p1,getid(obj));
     incpc(obj);
     break;
   case BIRTH:
     load(obj,p1,getbirth(obj));
     incpc(obj);
     break;
   case TIC:
     load(obj,p1,tic); //fix: puo' andare in overflow
                       //dividere in due posizioni HI-LO ?
     incpc(obj);
     break;
   case RAND:
     load(obj,p1,rand()%p2);
     incpc(obj);
     break;

   default:
     error(obj,DATA,ERROR,1,"steprun, cmd not exist");
     break;
 }
 //passaggio del tempo
 tic++;
 decerg(obj);
 return OK;
}
