 /*
   Assemblatore Ropeland
   
   @TITLE Rope Land
   @AUTHOR R.Corda
   @VERSION v0.2.22042000

   Copyright (c) 2000, Roberto Corda (roberto.corda@inferentia.it)
   All rights reserved.

   This file is subject to the terms of the GNU General Public License as
   published by the Free Software Foundation.  A copy of this license is
   included with this software distribution in the file COPYING.  If you
   do not have a copy, you may obtain a copy by writing to the Free
   Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "defs.h"
#include "rope.h"
#include "ropeasm.h"

#define MAXTOKENLEN 50
#define START "begin"
#define STOP "end"
#define SEPARATOR ";"
#define COPYLEFT "ACME_inc"
#define MAXPRG 10

int lexpos;
char token[MAXTOKENLEN];

char *close(void);
int insert(char c);


void _printcmd(tcmd *cmd){
  int i;

  printf("%s ",rlmname[cmd->rlm]);
  for (i = 0; i<npar[cmd->rlm]; i++) {
      if (cmd->ctrl[i] == VAL) printf("#%d ",cmd->p[i]);
      else printf("%d ",cmd->p[i]);
  }
  printf("\n\r");
}

void printallcmd(tcmd *cmd, int n){
  int i;

  for (i=0; i<n;i++){
    _printcmd(cmd+i);
  }
}
// estrae un token dal file f
// token e' qualsiasi cosa compreso tra due separatori
// i separatori sono (SPAZIO, TAB, RETURN, LINEFEED)
// fix: definizione troppo vaga di token
char *lex(FILE *f) {
 char c;
 char *out;

 while(
       (
        ((c = fgetc(f)) == '\n' ) ||
        (c == ' ') ||
        (c == '\r') ||
        (c == '\t')
       ) &&
        (!feof(f))
      )
 ;

 if (!feof(f)) insert(c);
 while(
       (!(
        ((c = fgetc(f)) == '\n' ) ||
        (c == ' ') ||
        (c == '\r') ||
        (c == '\t')
       ) ) &&
        (!feof(f))
      )
   if (!insert(c)) {
     out = close();
     printf("ERROR: token < %s > too long \n\r", out);
     free(out);
     fclose(f);
     exit(0);
   }
       ;

 out = close();

 return out;
}

int insert(char c){
  token[lexpos] = (char)c;
  if (++lexpos > MAXTOKENLEN - 1)
    return FALSE;
  else
    return TRUE;
}

char *close(void) {
  char *out;

  token[lexpos] = 0;
  out = strdup(token);
  lexpos = 0;
  token[0] = 0;
  return out;
}

// restituisce il  numero del comando mnemonico
// o 0 (zero) se non esiste
int mnemonic(char *ltoken) {
  int i;

  for(i=1;i<POOLCMD;i++)
    if(!strcmp(rlmname[i],ltoken))
      return i;
  return 0;
}


void getpar(char *ltoken,int *c,int *p) {
  //se e' un numero #value, aka non e' un puntatore
  if (ltoken[0] == ctrlname[VAL][0]){
    *c = VAL;
    *p = atoi(ltoken+1);
  } else {
    *c = PNT;
    *p = atoi(ltoken);
  }
}

// errori della funzione loadsource
int _loaderror(int cond, char *comment, char *fname,FILE *f,
               char *ltoken,int line) {
 if (cond) {
      printf("ERROR: %s in <%s>, tok:%s lin:%d\n\r",comment,fname,ltoken,line);
      if (ltoken != NULL)
        free(ltoken);
      fclose(f);
      return TRUE;
    }
 else return FALSE;
}


// carica e stampa elementi opzionali (coppie del tipo "@QUALCOSA pippo")
char *loadoptio(char *optio,char *ltoken, FILE *f) {

  if (!strcmp(ltoken,optio)) {
    // if (debug())
    printf("%s: ",ltoken);
    free(ltoken);
    ltoken = lex(f);
    // if (debug())
    printf("%s\n\r",ltoken);
    free(ltoken);
    ltoken = lex(f);
  }
 return ltoken;
}

//-----------------------------------------------------------
// legge i file di tipo .asm (sorgenti ling.macchina ropeland)
int loadsource(char *fname,tcmd *prg) {
    FILE *f = NULL;
    char *ltoken;
    int cmdnum = 0;
    int i,p,ctrl;

    //init lexical analyzer
    lexpos = 0;
    token[0] = 0;
    cmdnum = 0;

    f = fopen(fname,"r");

    ltoken =strdup(COPYLEFT);
    if (_loaderror((f == NULL),"loadsource, source file dosn't exist",
                fname,f,ltoken,cmdnum))
      return FALSE;
    free(ltoken);

    ltoken = lex(f);

    // copyright
    if (_loaderror(strcmp(ltoken,COPYLEFT),
      "loadsource, not a ropeland file",fname,f,ltoken,cmdnum))
      return FALSE;

    free(ltoken);

    ltoken = lex(f);
    ltoken = loadoptio("@TITLE",ltoken,f);
    ltoken = loadoptio("@AUTHOR",ltoken,f);
    ltoken = loadoptio("@VERSION",ltoken,f);

    // guarda se c'e' il token START
    if (_loaderror(strcmp(ltoken,START),
      "loadsource, source not started",fname,f,ltoken,cmdnum))
      return FALSE;

    free(ltoken);

    // carica i comandi
    ltoken = lex(f);
    while (strcmp(ltoken,STOP) && (!feof(f))){
      if (_loaderror(cmdnum >= MAXCMD,
        "loadsource, too many commands",fname,f,ltoken,cmdnum))
        return FALSE;

      // prende il valore associato al nome mnemonico
      prg[cmdnum].rlm = mnemonic(ltoken);

      if (_loaderror(! prg[cmdnum].rlm,"loadsource, not a command",
          fname,f,ltoken,cmdnum))
          return FALSE;

      free(ltoken);

      //prende i parametri
      for (i=0; i< npar[prg[cmdnum].rlm]; i++) {
        ltoken = lex(f);
        getpar(ltoken,&ctrl,&p);
        prg[cmdnum].ctrl[i] = ctrl;
        prg[cmdnum].p[i] = p;
        free(ltoken);
      }

      //fix: del this
      //if (debug)
      //_printcmd(prg + cmdnum);
      cmdnum++;
      ltoken = lex(f);
    }

    // controlla la presenza del token STOP
    if (_loaderror(strcmp(ltoken,STOP),
      "loadsource, program not ended",fname,f,ltoken,cmdnum))
      return FALSE;

    free(ltoken);
    fclose(f);
    return cmdnum;
}

//-----------------------------------------------------------
// legge i file di tipo .inc (stato iniziale)
// e inserisce oggetti in ropeland
// fix: per adesso e solo una copia di load source
int loadinc(char *fname) {
    FILE *f = NULL;
    char *ltoken;
    int prgnum = 0;   // numero di programmi .asm caricati
    int istnum[MAXPRG]; // numero istanze per ogni programma
    int curist;
    int fi,mi,ty,er,x,y,we; //parametri addobj
    int i,index,sum;
                                     // fix: ...
    tcmd *t[MAXPRG];                 // serve per load oggetti
    int n[MAXPRG];                   // num cmd oggetti caricati


    // matrice dove vengono caricati i programmi dall'esterno

    for (i= 0; i<MAXPRG; i++) {
      t[i] = malloc(sizeof(tcmd)*MAXCMD);
    }
    //init lexical analyzer
    lexpos = 0;
    token[0] = 0;

    f = fopen(fname,"r");

    ltoken =strdup(COPYLEFT);
    if (_loaderror((f == NULL),"loadinc, source file dosn't exist",
                fname,f,ltoken,prgnum))
      return FALSE;
    free(ltoken);

    ltoken = lex(f);

    // copyright
    if (_loaderror(strcmp(ltoken,COPYLEFT),
      "loadinc, not a ropeland file",fname,f,ltoken,prgnum))
      return FALSE;

    free(ltoken);

    ltoken = lex(f);
    ltoken = loadoptio("@TITLE",ltoken,f);
    ltoken = loadoptio("@AUTHOR",ltoken,f);
    ltoken = loadoptio("@VERSION",ltoken,f);

    // guarda se c'e' il token START
    if (_loaderror(strcmp(ltoken,START),
      "loadinc, .inc source not started",fname,f,ltoken,prgnum))
      return FALSE;

    free(ltoken);
    ltoken = lex(f);
    do {
       if (_loaderror(prgnum >= MAXPRG,
         "loadinc, too many programs",fname,f,ltoken,prgnum))
         return FALSE;
       free(ltoken);

       //file name file .asm
       ltoken = lex(f);
       //fix: if (debug())
         printf("loading: %s\n\r",ltoken);
       // carica sorgenti
       if(!(n[prgnum] = loadsource(ltoken,t[prgnum])))
         return FALSE;
       if (debug()) {
        printallcmd(t[prgnum],n[prgnum]);
        printf("--------------other source---------------\n\r");
       }
       free(ltoken);

       // number:
       ltoken = lex(f);
       free(ltoken);
       // value number
       ltoken = lex(f);
       istnum[prgnum] = atoi(ltoken);
       free(ltoken);
       // file
       ltoken = lex(f);


         prgnum++;
      } while (strcmp(ltoken, "("));

    // valori parametri programmi
    curist = 0;
    while (strcmp(ltoken,STOP) && (!feof(f))){
      //carica valori dentro parentesi usando numeroistanze[]
      free(ltoken);
      //father_id
      ltoken = lex(f);
      free(ltoken);
      //value father_id
      ltoken = lex(f);
      if (strcmp("GOD",ltoken))
        fi = atoi(ltoken);
      else
        fi = GOD;
      free(ltoken);
      // mother_id
      ltoken = lex(f);
      free(ltoken);
      // value mother_id
      ltoken =  lex(f);
      if (strcmp("GOD",ltoken))
        mi = atoi(ltoken);
      else
        mi = GOD;
      free(ltoken);
      // tipe
      ltoken = lex(f);
      free(ltoken);
      // value type
      ltoken = lex(f);
      if (!strcmp(ltoken,"LAMB")) {
        ty = LAMB;
      } else if (!strcmp(ltoken,"ROCK")) {
        ty = ROCK;
      } else if (!strcmp(ltoken,"LION")) {
        ty = LION;
      } else if (!strcmp(ltoken,"TREE")) {
        ty = TREE;
      } else
        ty = atoi(ltoken);
      free(ltoken);
      // erg
      ltoken = lex(f);
      free(ltoken);
      // value erg
      ltoken = lex(f);
      er = atoi(ltoken);
      free(ltoken);
      // weight
      ltoken = lex(f);
      free(ltoken);
      // value weight
      ltoken = lex(f);
      we = atoi(ltoken);
      free(ltoken);
      // x
      ltoken = lex(f);
      free(ltoken);
      // value x
      ltoken = lex(f);
      x = atoi(ltoken);
      free(ltoken);
      // y
      ltoken = lex(f);
      free(ltoken);
      // value y
      ltoken = lex(f);
      y = atoi(ltoken);
      free(ltoken);
      // ")"
      ltoken = lex(f);
      free(ltoken);

      // le ROCK non hanno associato un programma
      if (ty != ROCK) {
      // indice di t[..] e n[..]
      curist++;
      for (index=0,sum = istnum[0];
        curist > sum;
        sum += istnum[++index]);

        addobj(fi,ty,er,we,x,y,t[index],n[index]);
      } else  // aggiunge una ROCK
       addobj(fi,ty,er,we,x,y,NULL,0);

      // "("
      ltoken = lex(f);
    }

    // controlla la presenza del token STOP
    if (_loaderror(strcmp(ltoken,STOP),
      "loadinc, program not ended",fname,f,ltoken,prgnum))
      return FALSE;

    free(ltoken);

    for (i= 0; i<MAXPRG; i++) {
      free(t[i]);
    }
    fclose(f);
    return prgnum;
}

// fix: pare non salvi correttamente (oppure load)
// fix: quindi non lo utilizzo
int savepcode(char *fname, tcmd *prg, int *cmdnum) {
    FILE *f = NULL;
    int i,numread;
    char *cleft;

    cleft = strdup(COPYLEFT);

    f = fopen(fname,"w");
    if (_loaderror(f == NULL,"savepcode, on writing",
                  fname,f,cleft,*cmdnum))
      return FALSE;

    fwrite(cleft,sizeof(char)*strlen(COPYLEFT),1,f);
    fwrite(cmdnum, sizeof(int), 1, f);


    //fix: in questo modo la lunghezza del programma e' fissata
    numread = fwrite(prg, sizeof(tcmd), *cmdnum, f);

    if(_loaderror(numread != *cmdnum,
               "loadpcode, write error",fname,f,cleft,numread))
      return FALSE;
    fclose(f);
    return TRUE;
}

// fix: pare non carichi correttamente, non utilizzato
int loadpcode(char *fname, tcmd *prg) {
    FILE *f = NULL;
    int i,numread;
    int len;
    int cmdnum=0;
    char *cleft;

    len = strlen(COPYLEFT);
    cleft = malloc(sizeof(char)*len+1);
    cleft[0] = 0;

    f = fopen(fname,"r");
    if (_loaderror(f == NULL,"loadpcode, file doesn't exist",
                  fname,f,cleft,cmdnum))
      return FALSE;

    fread(cleft,sizeof(char)*len,1,f);
    cleft[len]=0;
    if (_loaderror(strcmp(cleft,COPYLEFT),"loadpcode, not a RopeLand file",
                  fname,f,cleft,cmdnum))
      return FALSE;

    fread(&cmdnum, sizeof(int), 1, f);
    //fix: in questo modo la lunghezza del file e' fissata
    numread= fread(prg, sizeof(tcmd), cmdnum, f);
    if(_loaderror(numread != cmdnum,
               "loadpcode, read error",fname,f,cleft,numread))
      return FALSE;

    fclose(f);
    return cmdnum;
}
